#lang scheme
(require "test.ss")

;; Combinators

;; ----------------------------------------

;; no-more : next
(define (no-more success-k fail-k)
  (fail-k))

;; or-k : next next -> (success fail -> alpha)
(define ((or-k first second) success-k fail-k)
  (local [(define (make-fst-success-k success-k fail-k)
            (lambda (val stream new-next)
              (success-k val stream
                         (lambda (success-k fail-k)
                           (new-next (make-fst-success-k success-k fail-k)
                                     (make-fst-fail-k success-k fail-k))))))
          (define (make-fst-fail-k success-k fail-k)
            (lambda ()
              (second success-k fail-k)))]
    (first (make-fst-success-k success-k fail-k)
           (make-fst-fail-k success-k fail-k))))

;; and-k : next (stream -> next) -> (success fail -> alpha)
(define ((and-k first second combine) success-k fail-k)
  (local [(define (make-fst-success-k success-k fail-k)
            (lambda (fst-vals stream fst-next)
              ((second stream)
               (make-snd-success-k fst-vals fst-next success-k fail-k)
               (lambda ()
                 (fst-next (make-fst-success-k success-k fail-k)
                           fail-k)))))
          (define (make-snd-success-k fst-vals fst-next success-k fail-k)
            (lambda (snd-vals stream snd-next)
              (success-k 
               (combine fst-vals snd-vals)
               stream
               (lambda (success-k fail-k)
                 (snd-next 
                  (make-snd-success-k fst-vals fst-next success-k fail-k)
                  (lambda ()
                    (fst-next (make-fst-success-k success-k fail-k)
                              fail-k)))))))]
    (first (make-fst-success-k success-k fail-k)
           fail-k)))

;; wrap-success-k : next (any -> any) -> success fail -> allpha
(define (wrap-success-k try val->val)
  (local [(define (make-my-success-k success-k fail-k)
            (lambda (val stream next)
              (success-k (val->val val) stream
                         (lambda (success-k fail-k)
                           (next (make-my-success-k success-k fail-k)
                                 fail-k)))))]
    (lambda (success-k fail-k)
      (try (make-my-success-k success-k fail-k) fail-k))))

;; ----------------------------------------

(define table
  `([main [("apple") ,(lambda () 'apple)]
          [("banana") ,(lambda () 'banana)]
          [(fruit-salad "!") ,(lambda (x)
                                        `(,x !))]]
    [fruit-salad [(fruit fruit) ,(lambda (x y)
                                   (list x y))]
                 [(fruit fruit fruit) ,(lambda (x y z)
                                         (list x y z))]]
    [fruit [("orange") ,(lambda () 'orange)]
           [("grape") ,(lambda () 'grape)]]))

;; parse : sym table list-of-string -> next
(define (parse what table stream)
  (local [(define defn (assq what table))]
    (if defn
        (parse-alts (rest defn) table stream)
        (error 'parse "no definition for ~a" what))))

(define (ok v stream next) (cons v stream))
(define (nope) #f)

(test ((parse 'main table '("apple")) ok nope) (cons 'apple '()))
(test ((parse 'fruit-salad table '("orange" "orange")) ok nope) 
      (cons '(orange orange) '()))
(test ((parse 'main table '("orange" "orange" "!")) ok nope)
      (cons '((orange orange) !) '()))
(test ((parse 'main table '("orange" "orange" "orange" "!")) ok nope)
      (cons '((orange orange orange) !) '()))
(test ((parse 'main table '("radish")) ok nope) #f)

;; parse-alts : list-of-case table list-of-string -> next
(define (parse-alts alts table stream)
  (if (null? alts)
      no-more
      (or-k (parse-case (first alts) table stream)
            (parse-alts (rest alts) table stream))))

(let ([alts (list
             `[("apple") ,(lambda () 'apple)]
             `[("grape") ,(lambda () 'grape)])])
  (test ((parse-alts alts table '("apple")) ok nope) (cons 'apple '()))
  (test ((parse-alts alts table '("grape")) ok nope) (cons 'grape '()))
  (test ((parse-alts alts table '("grape" "fish")) ok nope)
        (cons 'grape '("fish")))
  (test ((parse-alts alts table '("radish")) ok nope) #f))

;; parse-case : case table list-of-string -> next
(define (parse-case case table stream)
  (local [(define seq (first case))
          (define handler (second case))]
    (wrap-success-k
     (parse-sequence seq table stream)
     (lambda (vals) (apply handler vals)))))

(let ([case `[("apple") ,(lambda () 'apple)]])
  (test ((parse-case case table '("apple")) ok nope) (cons 'apple '()))
  (test ((parse-case case table '("grape")) ok nope) #f))


;; parse-sequence : sequence table list-of-string -> next
(define (parse-sequence seq table stream)
  (cond
    [(empty? seq) 
     (lambda (success-k fail-k)
       (success-k empty stream no-more))]
    [else 
     (and-k (parse-item (first seq) table stream)
            (lambda (stream) (parse-sequence (rest seq) table stream))
            append)]))

(let ([seq `("apple" "grape")])
  (test ((parse-sequence '("apple") table '("apple")) ok nope) 
        (cons empty '()))
  (test ((parse-sequence seq table '("apple" "grape")) ok nope)
        (cons empty '()))
  (test ((parse-sequence seq table '("apple" "radish")) ok nope) #f))
(let ([seq '(fruit fruit)])
  (test ((parse-sequence seq table '("orange" "grape")) ok nope)
        (cons '(orange grape) '()))
  (test ((parse-sequence seq table '("grape" "orange")) ok nope)
        (cons '(grape orange) '())))
(let ([seq '(fruit-salad "!")])
  (test ((parse-sequence seq table '("orange" "orange" "!")) ok nope)
        (cons '((orange orange)) '()))
  (test ((parse-sequence seq table '("orange" "orange" "orange" "!")) ok nope)
        (cons '((orange orange orange)) '())))

;; parse-item : sequence table list-of-string -> next
(define (parse-item item table stream)
  (cond
    [(string? item)
     (if (and (cons? stream)
              (equal? item (first stream)))
         (lambda (success-k fail-k)
           (success-k empty (rest stream) no-more))
         no-more)]
    [else
     (wrap-success-k (parse item table stream)
                     list)]))

(test ((parse-item '"grape" table '("grape")) ok nope)
      (cons empty '()))
(test ((parse-item 'fruit table '("grape")) ok nope)
      (cons '(grape) '()))
(test ((parse-item 'fruit table '("radish")) ok nope) #f)

(run-tests)
